// $Header: /u/drspeech/repos/quicknet2/QN_ListStream.cc,v 1.11 2014/05/06 03:03:32 davidj Exp $

// QN_ListStream.cc

// -*- C++ -*-
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include "QN_intvec.h"
#include "QN_ListStream.h"
#include "QN_Logger.h"
#include "QN_utils.h"

////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
// QN_InFtrStream_List - base class for a list of files
//   To see an example derived class implementation of this
//   class, take a look at 'QN_listStreamSRI.cc'.
////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
QN_InFtrStream_List::QN_InFtrStream_List(int a_debug, 
			      const char* a_dbgname, FILE* a_file,
					 int a_indexed, QN_ListType a_list_type) 
    : log(a_debug, "QN_InFtrStream_List", a_dbgname), 
      debug(a_debug),
      dbgname(a_dbgname),
      file(a_file), 
      current_seg(-1), 
      current_frame(0), 
      frames_this_seg(0),
      frame_ftrs(QN_SIZET_BAD), 
      header_peeked(0),
      indexed(a_indexed), 
      list_type(a_list_type),
      n_segs(QN_SIZET_BAD), 
      n_rows(QN_SIZET_BAD), 
      indexlen(0),
      index(NULL)
{

    log.log(QN_LOG_PER_RUN, "Creating InFtrStream_List from file '%s', list type %s.",
	    QN_FILE2NAME(a_file), list_type_str());
    curr_filename[0] = '\0';	// Just to be safe.
    curr_segname[0] = '\0';	// Just to be safe.

    // Build a regular expression for parsing the line
    char errbuf[1024];
    int ec;
    switch(list_type)
    {
    case QN_LISTTYPE_PLAIN:
	ec = regcomp(&regexp, "^(.*)[ \t]*$", REG_ICASE|REG_EXTENDED);
	break;
    case QN_LISTTYPE_SCP:
	ec = regcomp(&regexp, "^([^=]+)=([^[]*)\\[([0-9]+),([0-9]+)][ \t]*$", REG_ICASE|REG_EXTENDED);
	break;
    case QN_LISTTYPE_KALDI:
	ec = regcomp(&regexp, "^([^ ]+) ([^:]+):([0-9]+)[ \t]*$", REG_ICASE|REG_EXTENDED);
	break;
    default:
	log.error("Invalid list type %d in constructor.", list_type);
    }
    if (ec!=0)
    {
	regerror(ec, &regexp, errbuf, sizeof(errbuf));
	log.error("Failed to compile regular expression - %s.", errbuf);
    }
}

/*
** This method MUST be called from within the derived class constructor
** in order to initialize all of the base class variables.
*/
void
QN_InFtrStream_List::init() {
    // If required, build an index of the file.
    if (indexed) {
	log.log(QN_LOG_PER_RUN, "Creating index.");
	// Allocate and build buffer
	indexlen = DEFAULT_INDEX_LEN;
	index = new IndexEntry[indexlen];
	build_index();
	log.log(QN_LOG_PER_RUN, "Indexed stream, n_segs=%lu n_rows=%lu.",
		n_segs, n_rows);
	if (rewind()==QN_BAD) {
	    log.error("Failed to seek to start of %s after "
		      " building index.", QN_FILE2NAME(file));
	}
    }
    // We need to know how many bytes per frame there are so we can
    // calculate the number of features per frame
    if (frame_ftrs == QN_SIZET_BAD) {
	// Need to peek at the first header to get the number of bytes.
	// nextseg() will call open_ftrfile() which will set frame_bytes.
	if(nextseg() == QN_SEGID_BAD) {
	    log.error("Failed to read first header of %s.", 
		      QN_FILE2NAME(file));
	}
	// set this so that the next call to nextseg() will do nothing
	header_peeked = 1;
    }
}

QN_InFtrStream_List::~QN_InFtrStream_List() {
    if (index) {
	delete [] index;
    }
    regfree(&regexp);
}

/*
** Rewind the list stream.
*/
int QN_InFtrStream_List::rewind(void) {
    int ec;
    int ret;			// Return code

    ec = fseek(file, 0L, SEEK_SET);
    // Sometimes it is valid for rewind to fail (e.g. on pipes),
    // so return an error code rather than crash.
    if (ec!=0) {
	log.log(QN_LOG_PER_EPOCH, "Rewind failed for %s.", 
		QN_FILE2NAME(file));
	ret = QN_BAD;
    } else {
	current_seg = -1;
	current_frame = 0;
	log.log(QN_LOG_PER_EPOCH, "File %s rewound.", 
		QN_FILE2NAME(file));
	ret = QN_OK;
    }
    return ret;    
}

/*
** Increments current_seg and sets current_frame
** Returns QN_SEGID_BAD if it can't move to the next segment (i.e. EOF)
*/
QN_SegID QN_InFtrStream_List::nextseg() {
    QN_SegID id;		// SegID for next segment;

    if(header_peeked){
	header_peeked=0;
	return QN_SEGID_UNKNOWN;
    }

    current_seg++;

    if(indexed && static_cast<size_t>(current_seg) >= n_segs) // check for EOF
	return QN_SEGID_BAD;

    id = parse_listline();
    if (id == QN_SEGID_BAD)	// Handle EOF
	return QN_SEGID_BAD;

    // Read the next header (else EOF)
    open_ftrfile();
    current_frame = 0;

    return id;
}

/*
** Return the number of frames in the given segment, or all segments.
*/
size_t QN_InFtrStream_List::num_frames(size_t segno /* = QN_ALL */) {
    size_t ret;			// Return value.

    if (!indexed) {		// If not indexed, do not know no. of frames.
	ret = QN_SIZET_BAD;
    } else {
	if (segno == QN_ALL) {  // QN_ALL means do for whole stream.
	    ret = n_rows;
	} else {
	    // Look up number of frames in segment index.
	    assert (segno<n_segs);
	    ret = index[segno+1].segind - index[segno].segind; 
	}
    }
    return ret;
}

/*
** Will put the current segment number into segnop and the current
** frame number into framenop. If we are not at a valid position in
** the stream (i.e. current_seg == -1) then segnop and framenop get
** QN_SIZET_BAD.
** Returns QN_OK if indexed, QN_BAD otherwise.
*/
int QN_InFtrStream_List::get_pos(size_t* segnop, size_t* framenop) {
    int ret;			// Return value
    size_t segno;		// The segno value returned.
    size_t frameno;		// The frameno value returned.

    if (!indexed) {
	ret = QN_BAD;
    } else {
	if (current_seg==-1) {	// we are not at a valid position yet
	    segno = QN_SIZET_BAD;
	    frameno = QN_SIZET_BAD;
	} else {		// we are at a valid position
	    segno = static_cast<size_t>(current_seg);
	    frameno = static_cast<size_t>(current_frame);
	    assert(current_frame>=0);
	}
	if (segnop!=NULL)
	    *segnop = segno;
	if (framenop!=NULL)
	    *framenop = frameno;
	log.log(QN_LOG_PER_SENT, "Currently at sentence %lu, frame %lu.",
		static_cast<unsigned long>(segno), 
		static_cast<unsigned long>(frameno));
	ret = QN_OK;
    }
    return ret;
}

/*
** This function will position the stream at the beginning of the frame 
** given by frameno in the segment given in segno, such that the next
** call to read_ftrs() will begin with frame number frameno. The value of
** frameno is relative to the segment, not the whole stream, thus frameno
** should never be greater than or equal to the number of frames in segno.
** Assuming that frameno starts from 0.
** Returns QN_SEGID_UNKNOWN if indexed, QN_SEGID_BAD otherwise.
*/
QN_SegID QN_InFtrStream_List::set_pos(size_t segno, size_t frameno) {
    int ret;			// Return value.
    QN_SegID id;		// Segment ID.

    assert(segno < n_segs);	// Check we do not seek past end of file
    if (!indexed) {
	id = QN_SEGID_BAD;
    } else {
	// need to check to see that frameno is < number of frames in segno
	if(frameno > index[segno+1].segind - index[segno].segind) {
	    log.error("Seek beyond end of segment %lu.",
		     static_cast<unsigned long>(segno));
	}
	// jump to the correct location in the list of files
	long offset = index[segno].fpos;
	if (fseek(file, offset, SEEK_SET)!=0) {
	  log.error("Seek failed in file "
		    "'%s', offset=%li - file problem?",
		    QN_FILE2NAME(file), offset);
	}

	// read the line from the list file.
	id = parse_listline();
	
	if (id == QN_SEGID_BAD)
	{
	    log.error("Error reading file name from list "
		      "'%s'.",QN_FILE2NAME(file));
	}

	// read the header
	open_ftrfile();
	// Read in enough features to position the feat file at the 
	// desired location (i.e. the beginning of frame 'frameno').
	// If the frameno is 0, then we don't have to read any frames.
	if(frameno > 0)
	{
	    size_t i;
	    float *tmpftrs = new float[frame_ftrs];

	    for (i = 0; i<frameno; i++)
		read_ftrs(1,tmpftrs);
	    delete [] tmpftrs;
	}
	current_frame = frameno;
	current_seg = segno;
	log.log(QN_LOG_PER_SENT, "Seek to segment %lu, frame %lu.",
		static_cast<unsigned long>(segno), 
		static_cast<unsigned long>(frameno));

    }
    return id;
}


void
QN_InFtrStream_List::build_index()
{
    // The starting frame number.
    size_t frame = 0;	
    // Count of the seg number
    size_t seg = 0;
    // Space for index allocated in constructor
    index[seg].segind = (QNUInt32) frame;
    // Move to beginning of list of files
    if(rewind() != QN_OK){
	log.error("failed to build index for %s.",QN_FILE2NAME(file));
    }
    index[seg].fpos = 0L;

    // Get the name of each of the files in the list of files pointed to by
    // FILE* file and count how many frames there are in each one.
    // (space for curr_filename of size QN_MAXPATHLEN allocated in constructor)
    while(parse_listline()!= QN_SEGID_BAD) {
	// open_ftrfile() uses curr_filename. It should set the value
	// for 'frames_this_seg'. This is a virtual method and must be 
	// filled in by any subclass.
	open_ftrfile();

	// This is the beginning frame number for
	// the next segment.
	frame += frames_this_seg;
	++seg;
	// Here we increase the space allocated for the index if necessary
	if (seg>=indexlen) {
	    IndexEntry* new_index;	// Temp pointer to the new index
	    size_t new_indexlen;	// Temp store for lenght of new index

	    // Create a new index, including a copy of the old index
	    new_indexlen = indexlen * 2;
	    new_index =  new IndexEntry[new_indexlen];
	    for (size_t i=0; i<indexlen; i++)
		new_index[i] = index[i];
	    delete[] index;
	    index = new_index;
	    indexlen = new_indexlen;
	}
	index[seg].segind = static_cast<QNUInt32>(frame);
	index[seg].fpos = ftell(file);
    }
    n_rows = frame;		// total number of frames in all files
    n_segs = seg;		// total number of utterances
}

// Read in a line from the list file in the appropriate format.
// Sets curr_segname, curr_filename, read_offset, read_count.
// Returns QN_SEGID_BAD at end of file.
QN_SegID
QN_InFtrStream_List::parse_listline()
{
    // a line in the list file needs to have a path name, an utterance ID and
    // maybe some offset info.  Let's assume an utterance ID and offset info
    // are, together, smaller  than the path name.
    enum { BUFLEN = QN_MAXPATHLEN * 2 };

    char buf[BUFLEN];
    char* p;
    long start_offset;		// Start offset of file.
    long end_offset;		// End offset of file.

    p = fgets(buf,sizeof(buf),file);
    if (p==NULL)
    {
	if (feof(file))
	    return QN_SEGID_BAD;	// EOF.
	else
	{
	    log.error("Failed to read frame from list file"
		     " '%s', file_offset=%li - %s.",
		      QN_FILE2NAME(file), ftell(file), strerror(errno));
	}
    }
    else
    {
	// make sure there is a new line at the end of the line, and then get
        // rid of it.
	size_t buflen = strlen(buf);

	if (buf[buflen-1]!='\n')
	{
	    log.error("Missing trailing newline at end of file %s.\n",
		      QN_FILE2NAME(file));
	}
	else
	    buf[buflen-1] = '\0';

	// Up to 4 results are returned.
	regmatch_t match[5];	// Where we return our matches.
	int ec;			// Error code.

	ec = regexec(&regexp, buf, 5,  match, 0);
	if (ec!=0)
	{
	    log.error("Invalid format in list file %s: %s",
		      QN_FILE2NAME(file), buf);
	}
	switch(list_type)
	{
	case QN_LISTTYPE_PLAIN:
	    buf[match[0].rm_eo] = '\0';
	    strcpy(curr_filename, &buf[match[0].rm_so]);
	    curr_index1 = 0;
	    curr_index2 = 0;
	    break;
	case QN_LISTTYPE_SCP:
	    // Break up the strings.
	    // printf("substring offsets: %lu,%lu,%lu,%lu\n", (match[1]).rm_so, (match[2]).rm_so,(match[3]).rm_so,(match[4]).rm_so);
	    buf[match[1].rm_eo] = '\0';
	    buf[match[2].rm_eo] = '\0';
	    buf[match[3].rm_eo] = '\0';
	    buf[match[4].rm_eo] = '\0';
	    // Pull out the stuff we care about.
	    strcpy(curr_segname, &buf[match[1].rm_so]);
	    strcpy(curr_filename, &buf[match[2].rm_so]);
	    curr_index1 = strtol(&buf[match[3].rm_so], NULL, 10);
	    curr_index2 = strtol(&buf[match[4].rm_so], NULL, 10);
	    if (curr_index2<curr_index1)
	    {
		log.error("ending frame index is less than starting frame index for utterance %s.", curr_segname);
	    }
            // printf("%s=%s:[%li,%li]\n", curr_segname, curr_filename,
	    //        curr_index1, curr_index2);
	    break;
	case QN_LISTTYPE_KALDI:
	    buf[match[0].rm_eo] = '\0';
	    buf[match[1].rm_eo] = '\0';
	    buf[match[2].rm_eo] = '\0';
	    strcpy(curr_segname, &buf[match[0].rm_so]);
	    strcpy(curr_filename, &buf[match[1].rm_so]);
	    curr_index1 = strtol(&buf[match[2].rm_so], NULL, 10);
	    curr_index2 = 0;
	    log.error("KALDI lists not implemented yet.");
	    break;
	default:
	    log.error("Invalid list type.");
	}
    }
    return QN_SEGID_UNKNOWN;
}

const char*
QN_InFtrStream_List::list_type_str()
{
    const char* r = NULL;

    switch(list_type)
    {
    case QN_LISTTYPE_PLAIN:
	r = "plain";
	break;
    case QN_LISTTYPE_SCP:
	r = "HTK-SCP";
	break;
    case QN_LISTTYPE_KALDI:
	r = "Kaldi";
	break;
    default:
	log.error("Do not know string for unknown list type %i.", list_type);
    }
    return r;
}
