// $Header: /u/drspeech/repos/quicknet2/testsuite/pfile_scp_test.cc,v 1.1 2014/05/07 02:34:17 davidj Exp $
//
// test scp files and pfiles by reading identical files and comparing the
// results.  Note indexed streams are used.

#include <assert.h>
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <unistd.h>

#include "QuickNet.h"
#include "rtst.h"

void
pfile_scp_test(int debug, const char* dir, const char* pfile,
	       const char* scpfile)
{
    int ec;

    rtst_start("pfile_scp_test");

    ec = chdir(dir);
    rtst_assert(ec == 0);


    FILE* pfilefp = QN_open(pfile, "r");
    rtst_assert(pfilefp!=NULL);
    QN_InFtrLabStream_PFile pf(debug, "pfile", pfilefp, 1);

    FILE* scpfilefp = QN_open(scpfile, "r");
    rtst_assert(scpfilefp!=NULL);
    QN_InFtrStream_ListSCP scpf(debug, "scpfile", scpfilefp, 1);

    // Check the basic attributes of the files are the same
    rtst_assert(pf.num_segs()==scpf.num_segs());
    rtst_assert(pf.num_frames()==scpf.num_frames());
    rtst_assert(pf.num_ftrs()==scpf.num_ftrs());

    size_t num_segs = pf.num_segs();
    size_t num_ftrs = pf.num_ftrs();
    size_t num_tests = rtst_numtests;
    // Tests random seeks.
    {
	for (size_t i=0; i<num_tests; i++)
	{
	    size_t seg;

	    seg = rtst_urand_i32i32_i32(0, num_segs-1);
	    size_t pf_frames = pf.num_frames(seg);
	    size_t scpf_frames = scpf.num_frames(seg);
	    rtst_assert(pf_frames==scpf_frames);

	    size_t frame = rtst_urand_i32i32_i32(0, pf_frames-1);
	    size_t cnt = rtst_urand_i32i32_i32(1, pf_frames - frame);

	    size_t pf_segid = pf.set_pos(seg, frame);
	    rtst_assert(pf_segid!=QN_SEGID_BAD);
	    size_t scpf_segid = scpf.set_pos(seg, frame);
	    rtst_assert(scpf_segid!=QN_SEGID_BAD);

	    size_t size = cnt * num_ftrs;
	    float* pf_buf = rtst_padvec_new_vf(size);
	    float* scpf_buf = rtst_padvec_new_vf(size);
	    size_t pf_cnt = pf.read_ftrs(cnt, pf_buf);
	    size_t scp_cnt = scpf.read_ftrs(cnt, scpf_buf);
	    rtst_assert(pf_cnt==scp_cnt);
	    rtst_checkeq_vfvf(size, pf_buf, scpf_buf);
	    rtst_padvec_del(scpf_buf);
	    rtst_padvec_del(pf_buf);

	    size_t pf_segno, pf_frameno;
	    size_t scpf_segno, scpf_frameno;
	    int pf_ec = pf.get_pos(&pf_segno, &pf_frameno);
	    int scpf_ec = pf.get_pos(&scpf_segno, &scpf_frameno);
	    rtst_assert(pf_ec==0);
	    rtst_assert(scpf_ec==0);
	    rtst_assert(pf_segno==scpf_segno);
	    rtst_assert(pf_frameno==scpf_frameno);
	}
    }
    // Test sequential reads.
    int done = 0;
    size_t segno = 0;
    int pf_ec = pf.rewind();
    int scpf_ec = scpf.rewind();
    rtst_assert(pf_ec==QN_OK);
    rtst_assert(scpf_ec==QN_OK);
    while (!done)
    {
	size_t pf_frames = pf.num_frames(segno);
	size_t scpf_frames = scpf.num_frames(segno);
	rtst_assert(pf_frames==scpf_frames);

	size_t cnt = rtst_urand_i32i32_i32(1, pf_frames/10);
	size_t size = cnt * num_ftrs;
	float* pf_buf = rtst_padvec_new_vf(size);
	float* scpf_buf = rtst_padvec_new_vf(size);

	size_t pf_cnt = pf.read_ftrs(cnt, pf_buf);
	size_t scp_cnt = scpf.read_ftrs(cnt, scpf_buf);
	rtst_assert(pf_cnt==scp_cnt);
	rtst_checkeq_vfvf(pf_cnt * num_ftrs, pf_buf, scpf_buf);
	rtst_padvec_del(scpf_buf);
	rtst_padvec_del(pf_buf);

	if (pf_cnt < cnt)
	{
	    QN_SegID pf_segid;
	    QN_SegID scpf_segid;
	    pf_segid = pf.nextseg();
	    scpf_segid = scpf.nextseg();
	    // printf("segno=%lu pf_segid=%lu scpf_segid=%lu\n", (unsigned long) segno, (unsigned long) pf_segid, (unsigned long) scpf_segid);
	    segno++;
	    if (segno>=num_segs)
	    {
		rtst_assert(pf_segid==scpf_segid);
		rtst_assert(segno==num_segs);
		done = 1;
	    }
	}
    }
    QN_close(scpfilefp);
    QN_close(pfilefp);

    rtst_passed();
}


int
main(int argc, char* argv[])
{
    int arg;
    int debug = 0;		// Verbosity level

    arg = rtst_args(argc, argv);
    QN_logger = new QN_Logger_Simple(rtst_logfile, stderr,
				     "pfile_scp_test");

    assert(arg==argc-3);
    const char* dir = argv[arg++];
    const char* pfile = argv[arg++];
    const char* scpfile = argv[arg++];
    if (rtst_logfile!=NULL)
	debug = 99;
    pfile_scp_test(debug, dir, pfile, scpfile);
    rtst_exit();
}
